import numpy as np
import time
import threading
import logging
from datetime import datetime, timezone
from queue import Queue
from matrice.deploy.client_utils import get_image_data, FastAPIClientUtils, TritonClientUtils


class MatriceDeployClient:

    def __init__(self, session, deployment_id, auth_key):
        self.session = session
        self.rpc = self.session.rpc
        self.deployment_id = deployment_id
        self.auth_key = auth_key

        self.client = None
        self.client_number = 0
        self.clients = []
        self.instances_info = None
        self.index_to_category = {}

        self.log_prediction_info_queue = Queue()
        self.last_refresh_time = 0
        logging.info(f"Initializing MatriceDeployClient for deployment {deployment_id}")
        self.refresh_deployment_info()
        self.get_index_to_category()
        self.client_utils = FastAPIClientUtils() if self.server_type == "fastapi" else TritonClientUtils(self.connection_protocol, self.model_id)
        self.start_log_prediction_info_thread()

    def refresh_deployment_info(self):
        if (time.time() - self.last_refresh_time < 60) and self.instances_info:
            return
        self.last_refresh_time = time.time()

        logging.info(f"Refreshing deployment info for deployment {self.deployment_id}")
        response = self.rpc.get(f"/v1/deployment/get_deployment_without_auth_key/{self.deployment_id}")
        if response.get("success"):
            self.deployment_info = response["data"]
            self.model_id = self.deployment_info["_idModel"]
            self.model_type = self.deployment_info["modelType"]
            self.runningInstances = self.deployment_info["runningInstances"]
            self.instances_info = [
                {
                    "ip_address": instance["ipAddress"],
                    "port": instance["port"],
                    "instance_id": instance["modelDeployInstanceId"],
                }
                for instance in self.runningInstances
                if instance.get("deployed", False)
            ]
            self.server_type = self.deployment_info.get("serverType", "fastapi")
            self.connection_protocol = (
                "grpc" if "grpc" in self.server_type.lower() else "rest"
            )
            logging.info(f"Successfully refreshed deployment info. Found {len(self.instances_info)} running instances")
            return response
        else:
            error_msg = response.get("message", "Unknown error occurred")
            logging.error(f"Failed to refresh deployment info: {error_msg}")
            raise Exception(error_msg)

    def get_index_to_category(self):
        try:
            logging.info(f"Getting index to category mapping for model {self.model_id}")
            if self.model_type == "trained":
                url = f"/v1/model/model_train/{self.model_id}"
            elif self.model_type == "exported":
                url = f"/v1/model/get_model_train_by_export_id?exportId={self.model_id}"
            else:
                error_msg = f"Unsupported model type: {self.model_type}"
                logging.error(error_msg)
                raise Exception(error_msg)

            response = self.rpc.get(url)
            if not response.get("data"):
                error_msg = "No data returned from model train endpoint"
                logging.error(error_msg)
                raise Exception(error_msg)

            modelTrain_doc = response["data"]
            self.index_to_category = modelTrain_doc.get("indexToCat", {})
            logging.info(f"Successfully retrieved index to category mapping with {len(self.index_to_category)} categories")
            return self.index_to_category
        except Exception as e:
            error_msg = f"Failed to get index to category: {e}"
            logging.error(error_msg)
            raise Exception(error_msg)

    def get_prediction(
        self, image_path: str = None, image_bytes: bytes = None, image_url: str = None
    ):
        if not any([image_path, image_bytes, image_url]):
            error_msg = "Must provide one of: image_path, image_bytes, or image_url"
            logging.error(error_msg)
            raise ValueError(error_msg)

        logging.info("Getting prediction for image")
        image_data = get_image_data(image_path, image_bytes, image_url)
        client = self.get_client()
        start_time = time.time()

        results = self.client_utils.inference(client, image_data)
        latency = time.time() - start_time
        logging.debug(f"Prediction completed in {latency:.3f} seconds")

        self.log_prediction_info_queue.put((results, latency, datetime.now(timezone.utc).isoformat(), image_data, self.client["instance_id"]))

        return results

    async def get_prediction_async(
        self, image_path: str = None, image_bytes: bytes = None, image_url: str = None
    ):
        if not any([image_path, image_bytes, image_url]):
            error_msg = "Must provide one of: image_path, image_bytes, or image_url"
            logging.error(error_msg)
            raise ValueError(error_msg)

        logging.info("Getting async prediction for image")
        image_data = get_image_data(image_path, image_bytes, image_url)
        client = self.get_client()
        start_time = time.time()

        results = await self.client_utils.async_inference(client, image_data)
        latency = time.time() - start_time
        logging.debug(f"Async prediction completed in {latency:.3f} seconds")

        self.log_prediction_info_queue.put((results, latency, datetime.now(timezone.utc).isoformat(), image_data, self.client["instance_id"]))

        return results

    def setup_clients(self):
        if not self.instances_info:
            self.refresh_deployment_info()
        if not self.instances_info:
            error_msg = "No instances found"
            logging.error(error_msg)
            raise Exception(error_msg)

        logging.info("Setting up clients")
        self.clients = self.client_utils.setup_clients(self.instances_info)

        if not self.clients:
            error_msg = "Failed to setup any clients"
            logging.error(error_msg)
            raise Exception(error_msg)
        logging.info(f"Successfully set up {len(self.clients)} clients")

    def update_client(self):
        if not self.clients:
            self.setup_clients()

        if not self.clients:
            error_msg = "No clients available after setup"
            logging.error(error_msg)
            raise Exception(error_msg)

        if self.client_number >= len(self.clients):
            self.client_number = 0

        self.client = self.clients[self.client_number]
        self.client_number += 1
        logging.debug(f"Updated to client {self.client_number} of {len(self.clients)}")

    def get_client(self):
        self.update_client()
        return self.client

    def log_prediction_info(self, prediction, latency, request_time, image_data, instance_id): # TODO: add the triton structure handling and upload image
        try:
            logging.info("Logging prediction info")
            self.refresh_deployment_info()
            payload = {
                "prediction": (
                    prediction.tolist()
                    if isinstance(prediction, np.ndarray)
                    else prediction
                ),
                "latency": latency,
                "reqTime": request_time,
                "_idDeploymentInstance": instance_id,
                "isMLAssisted": False,
            }

            response = self.rpc.post(
                path=f"/v1/model_prediction/log_prediction_info/{self.deployment_id}",
                payload=payload,
            )
            logging.info("Successfully logged prediction info")
            return response
        except Exception as e:
            logging.warning(f"Failed to log prediction info: {e}")
            return None

    def log_prediction_info_thread(self):
        logging.info("Starting prediction info logging thread")
        while True:
            try:
                prediction, latency, request_time, image_data, instance_id = self.log_prediction_info_queue.get()
                self.log_prediction_info(prediction, latency, request_time, image_data, instance_id)
            except Exception as e:
                logging.error(f"Error in prediction info logging thread: {e}")
            finally:
                time.sleep(3)

    def start_log_prediction_info_thread(self):
        logging.info("Starting prediction info logging background thread")
        threading.Thread(target=self.log_prediction_info_thread, daemon=True).start()



"""Module for Session class handling project sessions."""
import os
from datetime import datetime, timedelta

from matrice.projects import Projects
from matrice.rpc import RPC
from matrice.utils import handle_response


class Session:
    """Class to manage sessions.

    Initialize a new session instance.

    Parameters
    ----------
    account_number : str
        The account number associated with the session.
    project_id : str, optional
        The ID of the project for this session.                          
    Example
    -------
    >>> session = Session(account_number="9625383462734064921642156")
    """

    def __init__(self, account_number, access_key, secret_key, project_id=None, project_name=None):
        self.rpc = RPC(access_key, secret_key, project_id=project_id)
        self.account_number = account_number
        self.access_key = access_key
        self.secret_key = secret_key
        self.last_refresh_time = datetime.now()
        self.project_id = project_id
        self.project_name = project_name
        if project_name and not project_id:
            self.project_id = self._get_project_id_by_name()
            self.refresh()
    
    def _get_project_id_by_name(self):
        path = f"/v1/project/get_project_by_name?name={self.project_name}"
        resp = self.rpc.get(path=path)
        if resp.get("success"):
            return resp.get("data")["_id"]
        else:
            raise Exception(f"Could not fetch project id from project name. Response: {resp}")
    
    def refresh(self):
        """
        Refresh the instance by reinstantiating it with the previous values.
        """
        # Check if two minutes have passed since the last refresh
        # if datetime.now() - self.last_refresh_time < timedelta(minutes=2):
        #     raise Exception("Refresh can only be called after two minutes since the last refresh.")
        
        # Capture the current state
        init_params = {
            'account_number': self.account_number,
            'access_key': self.access_key,
            'secret_key': self.secret_key,
            'project_id': self.project_id,
            'project_name': self.project_name
        }

        # Reinitialize the instance
        self.__init__(**init_params)


    def update(self, project_id):
        """
        Update the session with new project details.

        Parameters
        ----------
        project_id : str, optional
            The new ID of the project.


        Example
        -------
        >>> session.update(project_id="660b96fc019dd5321fd4f8c7")
        """
        self.project_id = project_id
        self.rpc = RPC(access_key=self.access_key , secret_key = self.secret_key, project_id=project_id)

    def close(self):
        """
        Close the current session by resetting the RPC and project details.

        Example
        -------
        >>> session.close()
        """
        self.rpc = None
        self.project_id = None

    def _create_project(self, project_name, input_type, output_type):
        """
        Create a new project with specified parameters.

        Parameters
        ----------
        project_name : str
            The name of the project to be created.
        input_type : str
            The type of input for the project (e.g., 'image').
        output_type : str
            The type of output for the project (e.g., 'classification').

        Returns
        -------
        tuple
            A tuple containing the response data, error message (if any).

        Example
        -------
        >>> response, error = session._create_project("New Project", "image", "classification")
        >>> if error:
        >>>     print(f"Error: {error}")
        >>> else:
        >>>     print(f"Project created with ID: {response['_id']}")
        """
        enabled_platforms = {
            "android": False,
            "ios": False,
            "tpu": False,
            "intelCPU": False,
            "gcloudGPU": False,
        }
        path = "/v1/project"
        headers = {"Content-Type": "application/json"}
        body = {
            "name": project_name,
            "inputType": input_type,
            "outputType": output_type,
            "enabledPlatforms": enabled_platforms,
            "accountType": "",
            "accountNumber": self.account_number,
        }
        resp = self.rpc.post(path=path, headers=headers, payload=body)
        if resp.get("success"):
            resp_data = resp.get("data")
            return resp_data, None
        else:
            error = resp.get("message")
            return None, error

    def create_classification_project(self, project_name):
        """
        Create a classification project.

        Parameters
        ----------
        project_name : str
            The name of the classification project to be created.

        Returns
        -------
        Projects
            An instance of the Projects class for the created project, or None if an error occurred.

        Example
        -------
        >>> project = session.create_classification_project("Image Classification Project")
        >>> if project:
        >>>     print(f"Created project: {project}")
        >>> else:
        >>>     print("Could not create project.")
        """
        resp, error = self._create_project(
            project_name=project_name, input_type="image", output_type="classification"
        )

        if error is not None:
            print(f"Could not create project: \n {error}")
        else:
            P = Projects(session=self, project_name=resp["name"])
            return P

    def create_detection_project(self, project_name):
        """
        Create a detection project.

        Parameters
        ----------
        project_name : str
            The name of the detection project to be created.

        Returns
        -------
        Projects
            An instance of the Projects class for the created project, or None if an error occurred.

        Example
        -------
        >>> project = session.create_detection_project("Object Detection Project")
        >>> if project:
        >>>     print(f"Created project: {project}")
        >>> else:
        >>>     print("Could not create project.")
        """
        resp, error = self._create_project(
            project_name=project_name, input_type="image", output_type="detection"
        )
        if error is not None:
            print(f"Could not create project: \n {error}")
        else:
            P = Projects(session=self, project_name=resp["name"])
            return P

    def create_segmentation_project(self, project_name):
        """
        Create a segmentation project.

        Parameters
        ----------
        project_name : str
            The name of the segmentation project to be created.

        Returns
        -------
        Projects
            An instance of the Projects class for the created project, or None if an error occurred.

        Example
        -------
        >>> project = session.create_segmentation_project("Instance Segmentation Project")
        >>> if project:
        >>>     print(f"Created project: {project}")
        >>> else:
        >>>     print("Could not create project.")
        """
        resp, error = self._create_project(
            project_name=project_name,
            input_type="image",
            output_type="instance_segmentation",
        )
        if error is not None:
            print(f"Could not create project: \n {error}")
        else:
            P = Projects(session=self, project_name=resp["name"])
            return P

   

    def list_projects(self, project_type="", page_size=10, page_number=0):
        """
        List projects based on the specified type.

        Parameters
        ----------
        project_type : str, optional
            The type of projects to list (e.g., 'classification', 'detection'). If empty, all projects are listed.

        Returns
        -------
        tuple
            A tuple containing the dictionary of projects and a message indicating the result of the fetch operation.

        Example
        -------
        >>> projects, message = session.list_projects("classification")
        >>> print(message)
        Projects fetched successfully
        >>> for project_name, project_instance in projects.items():
        >>>     print(project_name, project_instance)
        """
        path = "/v1/project/v2"
        if project_type != "":
            path += f"?items[0][field]=outputType&items[0][operator]=is&items[0][value]={project_type}&pageSize={page_size}&pageNumber={page_number}"
        else:
            path += f"?pageSize={page_size}&pageNumber={page_number}"

        resp = self.rpc.get(path=path)
        if resp.get("success"):
            projects_data = resp.get("data", {}).get("items", [])
            projects = {project["name"]: Projects(session=self, project_name=project["name"]) for project in projects_data}
            return projects, "Projects fetched successfully"
        else:
            message = resp.get("message")
            return {}, f"Failed to fetch projects: \n {message}"

    def get_project_type_summary(self):
        """
        Get the count of different types of projects.

        Returns
        -------
        tuple
            A tuple containing:
            - A dictionary with project types as keys and their counts as values if the request is successful.
            - An error message if the request fails.

        Example
        -------
        >>> project_summary, error = session.get_project_type_summary()
        >>> if error:
        >>>     print(f"Error: {error}")
        >>> else:
        >>>     print(f"Project type summary: {project_summary}")
        """
        path = "/v1/project/get_projects_count_by_type"
        resp = self.rpc.get(path=path)
        print(resp)
        data, error, message = handle_response(
            resp,
            "Successfully fetched project type summary",
            "An error occurred while fetching project type summary",
        )

        if error:
            return {}, error
        
        # Extract the project type summary from the response data
        project_type_summary = data

        # Return the project type summary as a dictionary
        return project_type_summary, None   


    def create_session(account_number, access_key, secret_key):
        """
        Create and initialize a new session with specified credentials.

        Parameters
        ----------
        account_number : str
            The account number to associate with the new session.
        access_key : str
            The access key for authentication.
        secret_key : str
            The secret key for authentication.

        Returns
        -------
        Session
            An instance of the Session class initialized with the given credentials.

        Example
        -------
        >>> session = create_session("9625383462734064921642156", "HREDGFXB6KI0TWH6UZEYR", "UY8LP0GQRKLSFPZAW1AUF")
        >>> print(session)
        <Session object at 0x...>
        """
        S = Session(account_number=account_number, access_key=access_key, secret_key=secret_key)
        return S
